/* 
 *  Mezzanine - an overhead visual object tracker.
 *  Copyright (C) Andrew Howard 2002
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
/***************************************************************************
 * Desc: Assign identities to blobs
 * Author: Andrew Howard
 * Date: 21 Apr 2002
 * CVS: $Id: ident.c,v 1.2 2002/05/03 04:55:52 inspectorg Exp $
 ***************************************************************************/

#include <assert.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "geom.h"
#include "opt.h"
#include "mezzanine.h"

//added spt2002
#include <stdio.h>

// Find the nearest blob
mezz_blob_t *ident_get_nearest(mezz_bloblist_t *bloblist, int class,
                                double max_r, double ox, double oy);

// Ident information
typedef struct
{
  mezz_objectlist_t *objectlist;     // List of objects in the current frame
} ident_t;

// The one and only instance of the identer
static ident_t *ident;


// Initialise the identer
int ident_init(mezz_mmap_t *mmap)
{
  int i;
  //char key[64];
  //const char *value;
  mezz_object_t *object;
  char *ident_with_number;  //eg "ident0", "ident4"

  ident = malloc(sizeof(ident_t));
  memset(ident, 0, sizeof(ident_t));
  // there are a couple of extra spaces here for ident_with_number
  ident_with_number = (char *)malloc(10*sizeof(char)); 
  // Use the mmap'ed object list
  ident->objectlist = &mmap->objectlist;

  // Load the object descriptions
  ident->objectlist->count = opt_get_int("ident", "object_count", 0);
  for (i = 0; i < ident->objectlist->count; i++)
  {
    object = ident->objectlist->objects + i;

    //make section string ie "ident[0]"
    sprintf(ident_with_number, "ident[%d]", i );

    //load the color classes for this object
    object->class[0] = opt_get_int(ident_with_number, "class[0]", 0);
    //second color is optional. by spt2002
    object->class[1] = opt_get_int(ident_with_number, "class[1]", NULL_COLOR);

    printf("object%d class[0]=%d class[1]=%d\n", i,object->class[0],object->class[1]); //debug

    object->max_disp = opt_get_double("ident", "max_disp", 0.10);
    object->max_sep = opt_get_double("ident", "max_sep", 0.15);
    object->max_missed = opt_get_int("ident", "max_missed", 30);
    object->missed = 1000;
  }
  
  return 0;
}


// Close the identer
void ident_term()
{
  // Clean up
  free(ident);
}


// Locate and identify objects from the list of blobs.
// This doesnt do any filtering on the pose.
mezz_objectlist_t *ident_update(mezz_bloblist_t *bloblist)
{
  int i;
  mezz_object_t *object;
  mezz_blob_t *ablob, *bblob;
  double mx, my; // position
  double dx, dy; // pose direction vector

  // for each of the objects in the objectlist,
  // look for an appropriate set of blobs.
  for (i = 0; i < ident->objectlist->count; i++)
  {
    object = ident->objectlist->objects + i;

    // number of missed frames (consecutive frames during which the
    // object was not seen).  Later, if the obect IS seen then this
    // is reset to zero.
    object->missed++;

    // if the object has NOT been missed too many times, then look for
    // object-blob of class[0] within a small radius ($mas_disp) from
    // where the object was last spotted.
    if (object->missed <= object->max_missed)
      ablob = ident_get_nearest(bloblist, object->class[0],
                                object->max_disp, object->px, object->py);

    // if the object HAS been missed too many times, then look over the
    // entire field for object-blob of color class[0].
    else
      ablob = ident_get_nearest(bloblist, object->class[0],
                                1e16, object->px, object->py);

    // if indent_get_nearest() found no suitable blob of color class[0],
    // then continue with next iteration (this object is not assigned
    // a new position).
    if (ablob == NULL)
      continue;

    // if a suitable blob WAS found, then look for a second, complement,
    // blob of color class[1] within distance $max_sep of the first blob.
    // But if this is a one-blob object we can skip this step
    if( object->class[1] != NULL_COLOR ){
      bblob = ident_get_nearest(bloblist, object->class[1],
				object->max_sep, ablob->wox, ablob->woy);
      
      // if no such second blob found, continue to next iteration
      // (this object is not assigned a new position).
      if (bblob == NULL)
	continue;

      // if two suitable blobs were found then assign a new position
      // to the object based on the position of the blobs.
      mx = (bblob->wox + ablob->wox)/ 2;
      my = (bblob->woy + ablob->woy)/ 2;
      dx = (bblob->wox - ablob->wox);
      dy = (bblob->woy - ablob->woy);
      //set object angle based on the direction
      object->pa = atan2(dy, dx)+M_PI_2; 
        // add pi/2 to get angle perpendicular to line connecting 2 blobs
      if( object->pa >= M_PI )
	  object->pa -= 2*M_PI;
	  //since we are adding an extra pi/2, "overflow" over pi is possible


      // Assign the blobs to this object
      ablob->object = i;
      bblob->object = i;
    }
    else{ // one-blob object case
      // assign a new position to the object based on the position 
      // of the blobs.
      mx = ablob->wox;
      my = ablob->woy;

      object->pa = 0; // it doesnt make sense to give an angle to a single
                      // blob object.

      // Assign the one blob to this object
      ablob->object = i;
    }

    // assign velocity to the object by finding difference between
    // last position and current.
    // velocity is scaled based on the number of missed frames.

    if (object->missed <= object->max_missed){
      object->vx = (mx - object->px) / object->missed;
      object->vy = (my - object->py) / object->missed;
    }
    // However, if this is a "relocated object" (ie we looked all over the image
    // because it was missed too manu times) then we don't assign a velocity
    // (because this velocity would be greater than the maximum displacement
    //  allows)
    else{
      object->vx = 0;
      object->vy = 0;
    }

    // reset missed counter
    object->missed = 0;
    object->px = mx;
    object->py = my;

  }

  return ident->objectlist;
}


// Find the nearest unassigned blob of color class $class.
// Only search within radius $max_r of point ($ox,$oy).
// Returns NULL if no blob satisfies these conditions.
mezz_blob_t *ident_get_nearest(mezz_bloblist_t *bloblist, int class,
                                double max_r, double ox, double oy)
{
  int i;
  double dx, dy, dr;
  double min_dr;
  mezz_blob_t *blob, *min_blob;

  min_dr = max_r;
  min_blob = NULL;
  
  // for each blob in the bloblist
  for (i = 0; i < bloblist->count; i++)
  {
    blob = bloblist->blobs + i;

    // if this blob is the wrong color class, continue to next iteration
    if (blob->class != class)
      continue;

    // if this blob is already assigned an object
    // then continue to next iteration
    if (blob->object >= 0)
      continue;

    // if we get here, then the blob is a legal candidate
    dx = blob->wox - ox;
    dy = blob->woy - oy;
    dr = sqrt(dx * dx + dy * dy);

    // test to see that the distance from blob to point (ox,oy) is
    // less than the maximum distance allowed.
    if (dr < min_dr)
    {
      min_blob = blob;
      min_dr = dr;
    }
  }
  
  return min_blob;
}



