#include "Robot.h"
#include "PotentialMap.h"
#include "RoboMath.h"
#include "VisionClient.h"
#include "CommServer.h"
#include "Ball.h"
#include "Config.h"
#include "ActionTable.h"

#include <time.h>

#define BUFSIZE 1024

VisionClient visionClient;
CommServer commServer;
Ball* ball;
Robot* robots;
Config config;

void initObjects(Robot* robots, Ball* ball);

Vector get_traj( Robot* robot, Vector* target );
void move_naive( Robot* robot, Vector* pos );

void kill_handler( int signum ){
  printf("program terminated\n");
  exit(0);
}


int main(int argc, char** argv) {
  Vector center(2100, 1350);
  Vector goal(0, 1350);
  //init
  robots = new Robot[8];
  ball = new Ball();
  initObjects(robots, ball);
  commServer.init(config.commPort);
  visionClient.init(config.visHost, config.visPort);
  int manualIterationStep = 0;
  int testMotors = 0;
  int naiveMotion = 0;
  config.verbose = 0;

  //parse commandline parameters
  for( int i= 1; i < argc; i++){
    if (i==1) // test must be first parameter
      if(!strcmp(argv[i],"--test"))
	testMotors = 1;    
    else if (!strcmp(argv[i],"--step"))
      manualIterationStep = 1;
    else if (!strcmp(argv[i],"--verbose"))
      config.verbose = 1;
    else if (!strcmp(argv[i],"--naive"))
      naiveMotion = 1;    
  }

  //setup signal handlers
  if( signal( SIGTERM, kill_handler ) == SIG_IGN )
    signal( SIGTERM, SIG_IGN );
  if( signal( SIGINT, kill_handler ) == SIG_IGN )
    signal( SIGINT, SIG_IGN );

  //////////////////////////////////////////
  if( testMotors ){
    double l_direction, r_direction;
    char c, *buf;
    buf = (char *)malloc(BUFSIZE);
    int i,read_size;
    size_t buf_size;
    buf_size = BUFSIZE;
    
    double mag;
    if (argc == 2)
      mag = 1;
    else{
      mag = strtod(argv[2],NULL); //assume second command line parameter was the motor power expressed as a float between [0,1]
    }
    printf("type a key command followed by the enter key\n");
    printf("'5' forward\n'2' backward\n'1' left\n'3' right\n");
    
    while(1) {
      read_size = getline(&buf,&buf_size , stdin);
      
      for(i=0; i<read_size-1; i++){
	c = buf[i];
	//choose direction
	switch(c){
	case '5': //forward
	  l_direction = mag;
	  r_direction = mag;
	  break;
	case '2': //backward
	  l_direction = -mag;
	  r_direction = -mag;
	  break;
	case '1': //left
	  l_direction = -mag;
	  r_direction = mag;
	  break;
	case '3': //right
	  l_direction = mag;
	  r_direction = -mag;
	  break;
	default:
	  l_direction = mag;
	  r_direction = mag;
	}
	printf( "(%f,%f)\n", l_direction, r_direction );
	
	for (int i = 0; i < 8; i++) {
	  robots[i].setMotors(r_direction, l_direction);
	}
	commServer.sendData(robots);
      }
    }
  }

  else { ///////////////////////////////////////advanced mode
    
    clock_t start, end;
    end = clock();
    
    while(1) {
      if(config.verbose){
	cout << "\n===============Next frame================\n";
	cout << "Requesting data...";
      }
      visionClient.getData(robots, ball);
      
      if(config.verbose){
	cout << "Got data.\n";
	ball->getPosition().print("  Ball position");
      }
      
      // assign movement commands
      for(int i = 0; i < config.numOurRobots; i++){
	printf("\n  Robot %d:\t", i);
	if(config.verbose){
	  robots[i].getOrientationVector().print("ori");
	  robots[i].getPosition().print("   pos");
	  robots[i].getMotors().print("   motors_state");
	}
	
	if ( naiveMotion )
	  move_naive( &(robots[i]), &(ball->getPosition()) );
	else{
	  Vector traj_to_goal = get_traj( &(robots[i]), &(goal) );
	  robots[i].move( ball->getPosition(),traj_to_goal);
	}
	robots[i].getMotors().print("   motors_set");
      }
      
      if(config.verbose){
	printf("\n-------------Send Data---------------\n");  
      }
      commServer.sendData(robots);
      
     
      start = end;
      end = clock();
      printf("time elapsed = %lu\n", end - start);
     
      

      // if step is on, wait for input before continuing
      if (manualIterationStep == 1) getchar();

    } //end loop
    
  } ///////////////////////////////////////end advanced mode

}





void initObjects(Robot* robots, Ball* ball)
{
   
  for(int i = 0; i < config.numRobots; i++)
    {
      //      robots[i].init(config.fieldDimensions, config.numCells);
      robots[i].setName(config.names[i]);
      robots[i].setID(config.ids[i]);
    }

  ball->setName(config.ballName);
}

//set robot to move to target position
void move_naive( Robot* robot, Vector* target ){
  double angle_to_target;
  Vector traj; //this is the direction we want to go in
  traj.setX( target->x() - robot->getPosition().x() );
  traj.setY( target->y() - robot->getPosition().y() );
  traj = traj.normalize();
  
  angle_to_target = robot->getOrientationVector().abs_angle_diff(traj);

  if(config.verbose){
    cout << "   trajectory to target " << traj.angle() << "\n";
    traj.print("   trajectory to target");
    cout << "   angle to target " << angle_to_target << "\n";
  }
  robot->setMotors(1,1);

//if we are basically facing the right direction
  if( angle_to_target < 1 )
    robot->setMotors(1,1);
  else{
    if (robot->getOrientationDouble() - traj.angle() < 0)
      robot->setMotors(0.3,-0.3);
    else 
      robot->setMotors(-0.3,0.3);
  }
}

Vector get_traj( Robot* robot, Vector* target ){
  return *target - robot->getPosition();
}
