/* 
 *  Mezzanine - an overhead visual object tracker.
 *  Copyright (C) Andrew Howard 2002
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
/***************************************************************************
 * Desc: eleague braodcasting program
 * Author: Stephen Tarzia
 * Date: 25 Dec 2004
 ***************************************************************************/

#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "mezz.h"
#include <sys/time.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <string.h>

#define BUF_SIZE 1024

// Set flag to 1 to force program to quit
static int quit = 0;

// socket descriptors
struct sockaddr_in dst, dst2;

// Handle quit signals
void sig_quit(int signum)
{
  quit = 1;
}

// Get the wall clock time in seconds, taken from mezzanine.c
double gettime()
{
  struct timeval time;
  gettimeofday(&time, NULL);
  return (double) time.tv_sec + (double) time.tv_usec / 1e6;
}

/* Sets up UPD broadcast socket.
 */
int openport(int *sk_fd, int port){

  int setreuse = 1;
  if (-1 == (*sk_fd = socket (AF_INET, SOCK_DGRAM, 0))){
    fprintf( stderr, "socket");
    return -1;
  }
  int on = !0;            
  if (-1 == setsockopt (*sk_fd, SOL_SOCKET, SO_BROADCAST, &on, sizeof (on))){
    fprintf( stderr, "BROADCAST");       
    return -1;
  }
  setreuse = 1;
  if ( setsockopt( *sk_fd, SOL_SOCKET, SO_REUSEADDR, 
		   &setreuse, sizeof (setreuse))
       == -1){
    perror ("setsockopt");
    return -1;
  }

  dst.sin_family = AF_INET;
  dst.sin_port = htons (port - 1);
  dst.sin_addr.s_addr = htonl (INADDR_ANY);
  bzero (&(dst.sin_zero), 8);
  if (-1 == bind (*sk_fd, (struct sockaddr *) &dst, sizeof (dst))){
    fprintf( stderr, "bind");
    return -1;
  }

  dst2.sin_family = AF_INET;
  dst2.sin_port = htons (port);
  dst2.sin_addr.s_addr = htonl (INADDR_BROADCAST);
  bzero (&(dst2.sin_zero), 8);

  return 0;
}

int dowrite( int socket_fd, char *buf, int size){
  if ((unsigned int)
      sendto (socket_fd, buf, size, 0,
              (struct sockaddr *)&dst2, sizeof (dst2)) 
      != size )
    {                                                                                
      perror ("sendto unable to send message\n");                      
    }   
  return 0;
}

// Main
int main(int argc, char **argv)
{
  int i;
  mezz_mmap_t *mmap;
  mezz_object_t *object;
  double last_time;
  int socket_fd, port = 0;
  char buf[BUF_SIZE];
  double x_scalingfactor = 2300; // 4600mm pitch length for legged league
  double y_scalingfactor = 2700; // 2700mm pitch width for legged league
  int debug = 0;

  printf("This program broadcasts the robot and ball data using a UDP socket.  \nThe transmission is asynchronous and will therefore FLOOD THE NETWORK,\nmaking your network administrator very unhappy.\n");
    
  //parse commandline parameters
  for( i= 1; i < argc; i++){
    if(!strcmp(argv[i],"-x"))
      x_scalingfactor = 0.5 * atoi( argv[i+1] );
    else if (!strcmp(argv[i],"-y"))
      y_scalingfactor = atoi( argv[i+1] );
    else if (!strcmp(argv[i],"-d"))
      debug = 1;
    else if (!strcmp(argv[i],"-p"))
      port = atoi( argv[i+1] );
  }

  // parse runtime parameters
  if(port == 0){
    fprintf( stderr, "please pass the port number as a parameter, -p [port]\n");
    return -1;
  }

  printf( "using port number %d\n", port);


  // Register signal handlers
  signal(SIGINT, sig_quit);
  signal(SIGQUIT, sig_quit);
  
  // Initialise the IPC
  if (mezz_init(0) < 0)
    return -1;
  mmap = mezz_mmap();

  // Initialize UDP broadcast socket
  openport( &socket_fd, port );
  if (socket_fd == -1 ) return -1;

  // Initialize the output buffer
  //buf = (char *)malloc( BUF_SIZE * sizeof(char) );
  int k;

  // Initialize time
  last_time = gettime();

  // the output loop
  while (!quit)
  {
    int robot_number = 0;
    double orientation;
    // Make sure the IPC is running before we do anything else
    mezz_wait_event();

    // line 1: number_objects frame_number time_difference
    k = snprintf(buf, BUF_SIZE, "%d %d %f\n", mmap->objectlist.count, 
		 mmap->count, mmap->time - last_time);
    // line 2: coordinates_of_camera_wrt_real_world (unimplemented)
    k += snprintf(buf+k, BUF_SIZE-k, "0.0 0.0 0.0\n");

    // line 3 through 10: object information
    for (i = 0; i < mmap->objectlist.count; i++){
    
      object = mmap->objectlist.objects + i; // the object for this output line

      if (object->class[1] == NULL_COLOR )
	k += snprintf(buf+k, BUF_SIZE-k, "1 ball "); // type of object
      else{
	k += snprintf(buf+k, BUF_SIZE-k, "0 blue%d ", robot_number);
	robot_number ++;
      }
      
      //      orientation = atan(tan(object->pa * (y_scalingfactor/x_scalingfactor)));
      orientation = object->pa;
      if( object->pa <= 0 ){
	if( object->pa <= M_PI_2 ) // first quadrant
	  ;
	else // second quadrant
	  orientation += M_PI;
      }else {
	if( object->pa >= -M_PI_2 ) // fourth quadrant
	  ;
	else //third quadrant
	  orientation -= M_PI;
      }
      k += snprintf(buf+k, BUF_SIZE-k, "Found %f %f 0.0 %f %f %f\n", 
		    /* x position in mm */
		    (object->px+1) * x_scalingfactor,
		    /* y position in mm */
		    //((-object->py)+0.5) * y_scalingfactor,
		    (object->py+0.5) * y_scalingfactor,
		    /* "scaling" the angle is a bit messy */
		    //atan2(tan(object->pa) * y_scalingfactor,x_scalingfactor),
		    //atan2(tan(object->pa * (y_scalingfactor/x_scalingfactor)),
		    orientation,
		    /* x component of velocity in mm/frame */
		    object->vx * x_scalingfactor,
		    /* y component of velocity in mm/frame */
		    object->vy * y_scalingfactor);
      
    }

    // output to stdout
    if(debug)
      printf( "%s", buf );
    
    // write to socket
    dowrite( socket_fd, buf, BUF_SIZE );

    last_time = mmap->time;
  }

  
  //free(buf);

  // Finalise the IPC
  mezz_term(0);
  
  return 0;
}

