/* 
 *  Mezzanine - an overhead visual object tracker.
 *  Copyright (C) Andrew Howard 2002
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
/***************************************************************************
 * Desc: eleeague output from mezzanine
 * Author: Stephen Tarzia
 * Date: 25 Dec 2004
 ***************************************************************************/

#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "mezz.h"
#include <sys/time.h>

#include <sys/types.h>
#include <string.h>

#define BUF_SIZE 1024


// Set flag to 1 to force program to quit
static int quit = 0;

// Handle quit signals
void sig_quit(int signum)
{
  quit = 1;
}

// Get the wall clock time in seconds, taken from mezzanine.c
double gettime()
{
  struct timeval time;
  gettimeofday(&time, NULL);
  return (double) time.tv_sec + (double) time.tv_usec / 1e6;
}

int dowrite( char *buf, int size){
  printf( "%s", buf );
  return 0;
}

// Main
int main(int argc, char **argv)
{
  int i;
  mezz_mmap_t *mmap;
  mezz_object_t *object;
  double last_time;
  char buf[BUF_SIZE];
  double x_scalingfactor = 2300; // 4600mm pitch length for legged league
  double y_scalingfactor = 2700; // 2700mm pitch width for legged league

  //parse commandline parameters
  for( i= 1; i < argc; i++){
    if(!strcmp(argv[i],"-x"))
      x_scalingfactor = 0.5 * atoi( argv[i+1] );
    else if (!strcmp(argv[i],"-y"))
      y_scalingfactor = atoi( argv[i+1] );
  }

  if( argc == 3 ){
    y_scalingfactor = atoi( argv[2] );
  }
    
  // Register signal handlers
  signal(SIGINT, sig_quit);
  signal(SIGQUIT, sig_quit);
  
  // Initialise the IPC
  if (mezz_init(0) < 0){
    return -1;
  }
  mmap = mezz_mmap();

  // Initialize the output buffer
  //buf = (char *)malloc( BUF_SIZE * sizeof(char) );
  int k;

  // Initialize time
  last_time = gettime();

  // the output loop
  while (!quit)
  {
    int robot_number = 0;
    double orientation;
    // Make sure the IPC is running before we do anything else
    mezz_wait_event();

    // line 1: number_objects frame_number time_difference
    k = snprintf(buf, BUF_SIZE, "%d %d %f\n", mmap->objectlist.count, 
		 mmap->count, mmap->time - last_time);
    // line 2: coordinates_of_camera_wrt_real_world (unimplemented)
    k += snprintf(buf+k, BUF_SIZE-k, "0.0 0.0 0.0\n");

    // line 3 through 10: object information
    for (i = 0; i < mmap->objectlist.count; i++){
    
      object = mmap->objectlist.objects + i; // the object for this output line

      // assume that all one-color objects are a ball
      if (object->class[1] == NULL_COLOR ){
	k += snprintf(buf+k, BUF_SIZE-k, "1 ball "); // type of object
      }
      // also assume that there is only one team, namely blue.  
      else{
	k += snprintf(buf+k, BUF_SIZE-k, "0 blue%d ", robot_number);
	robot_number ++;
      }
      
      //      orientation = atan(tan(object->pa * (y_scalingfactor/x_scalingfactor)));
      orientation = object->pa;
      if( object->pa <= 0 ){
	if( object->pa <= M_PI_2 ) // first quadrant
	  ;
	else // second quadrant
	  orientation += M_PI;
      }else {
	if( object->pa >= -M_PI_2 ) // fourth quadrant
	  ;
	else //third quadrant
	  orientation -= M_PI;
      }
      k += snprintf(buf+k, BUF_SIZE-k, "Found %f %f 0.0 %f %f %f\n", 
		    /* x position in mm */
		    (object->px+1) * x_scalingfactor,
		    /* y position in mm */
		    /* here we are inverting the y axis from mezzanine's up == positive
		       to uleague's down == positive */
		    (-object->py+0.5) * y_scalingfactor,
		    /* "scaling" the angle is a bit messy */
		    //atan2(tan(object->pa) * y_scalingfactor,x_scalingfactor),
		    //atan2(tan(object->pa * (y_scalingfactor/x_scalingfactor)),
		    orientation,
		    /* x component of velocity in mm/frame */
		    object->vx * x_scalingfactor,
		    /* y component of velocity in mm/frame */
		    object->vy * y_scalingfactor);
      
    }

    // write to output stream
    dowrite( buf, BUF_SIZE );

    last_time = mmap->time;
  }

  // Finalise the IPC
  mezz_term(0);
  
  return 0;
}

